from math import sqrt
class zmodn:
    @staticmethod
    def mk_init_sub(mod):
        """
        given a modulus, return a constructor that expects
        a number and sets the object's .number and .modulus attributes
        """
        def __init__sub(self, n):
            self.n = n % mod
            self.mod = mod
        return __init__sub

    @staticmethod
    def prime(k):  # brute force...
        for pp in [2] + list(range(3,int(sqrt(k))+1,2)):
            if k % pp == 0:
                return False
        return True
        
    def __init_subclass__(cls):
        """
        Abuse of __init__subclass__:
        Parses the name of the derived class, which is
        expected to match the regex /_[0-9]+$/,
        extracts the digits after the last underscore,
        forms a modulus from that number,
        and builds and injects a constructor method
        for the class of numbers modulo that modulus.
        If the modulus is prime, adds a division operator.
        """
        name = cls.__name__
        if not "_" in name:
           raise TypeError("no underscore: child class' name must end in underscore followed by modulus")
        u = name.rindex("_")
        clsmodstr = name[u+1:]
        if not clsmodstr.isdigit():
           raise TypeError("no modulus: child class' name must end in underscore followed by modulus")
        clsmod = int(clsmodstr)
        field = False
        if zmodn.prime(clsmod):
            cls.field = True
            # Fields have maximal unitgroup
            cls.__truediv__ = zmodn.__div__
            cls.__floordiv__ = zmodn.__div__
        else:
            cls.field = False
        cls.__init__ = zmodn.mk_init_sub(clsmod)   # inject constructor

    def __add__(self, other):
        return self.__class__(self.n + other.n)

    def __neg__(self):
        return self.__class__(-self.n)

    def __sub__(self, other):
        return self + (-other)

    def __mul__(self, other):
        return self.__class__(self.n*other.n)

    @staticmethod
    def ext_eucl(a,b):
        if b == 0:
            return (1,0)
        else:
            (q,r) = divmod(a,b)
            (s,t) = zmodn.ext_eucl(b,r)
            return (t, s - q *t)

    def inv(self):
        if self.n == 0:
            raise ZeroDivisionError
        inn = zmodn.ext_eucl(self.n, self.mod)[0]
        return self.__class__(inn)

    def __div__(self,other):
        return self * other.inv()

    def __pow__(self, exp):
        ret = self.__class__(1)
        b   = self.__class__(self.n)
        while (exp > 0):
            if exp % 2 == 1:
                ret = ret * b
            exp = exp // 2
            b = b * b
        return ret

    def __str__(self):
        return str(self.n) + " mod " + str(self.mod)

    def __eq__(self, other):
        if isinstance(other, int):  # zmodn_5(2) == 2 -> True
            return self.n == other % self.mod
        else:
            if isinstance(other, self.__class__):
                return self.n == other.n
            else:
                raise TypeError("incomparable elements in different rings")

    def __neq__(self,other):
        return not self == other
    
# Tests

try:
    class zmod12(zmodn): pass
except TypeError as e:
    # check for missing underscore
    assert "no underscore" in "".join(e.args)

try:
    class zmod_eleven(zmodn): pass
except TypeError as e:
    # check for missin mod
    assert "no modulus" in "".join(e.args) 

class zmod_12(zmodn): pass
try:
    two = zmod_12(2)
    print(two / two)
except Exception as e:
    assert("unsupported operand" in "".join(e.args))

assert not zmod_12.field

class field_13(zmodn): pass
two13 = field_13(2)
print(list(map(str, [two13**i for i in range(13)])))
assert field_13.field

class field_65537(zmodn): pass
five = field_65537(5)
print(list(map(str, [five**p  for p in range(65530, 65537)])))

class bad_rsa_29_times_31_is_899(zmodn): pass
three = bad_rsa_29_times_31_is_899(3)
assert (three**3 + three - three**0)**31 == (three**3 + three - three**0)
